
/* Multi Switch Decoder Sofware V2.6 (Stand 06.03.2021)
 * Ersteller P. Doggwiler
 * Datum: 07.05.2020
 * Software Beschrieb
   * Über einen Proportional Kanal oder Taster können bis zu 11 Ausgänge geschaltet werden. Über einen Counter wird der gewünschte Ausgang hochgezählt.
   * Erfolg nach 1.5 Sekunden kein neues Counter Signal, wird der Ausgang entsprechend invertiert und der Counter wieder auf 0 gesetzt.
   * Einsatz der Printplatte: 
   * Je nach Ausgbau und setzen des Jumpers kann über deine Trainstorschaltung max. 0.5A direkt geschaltet werden (was der Transitor hergib)
   * oder direkt ein Relais Modul aus dem Arduino Zubehör für grosse Lasten
   * Counter (Ausgang) hochzählen erfolgt wenn Signal > 1800 oder < 1200 ist
   * Je nach gestecktem Jumper wird eine externe Spannung überwacht, welche bei unterschreiten eines Wertes ein Ausgang schaltet
   * Spanunng Min => Variable countAnzMaxSpanMin Wert Anpassen
   * Spannungsüberprüfung => Variable maxCountSpannung  Wert Anpassen falls gewünscht
   * 
   * 
   * 
   * <<<<<<<<<<<<<<<<<<<<<<!!!!!!!! Hinweise zur Software Inbetriebnahme !!!!!!!!>>>>>>>>>>>>>>>>>>>>>>
   * 
   * - Auf dem Board müssen Jumper 17 und Jumper 18 auf GND oder VCC gesetzt werden. VCC ist aktiv geschaltet
   * - Wenn Spannungsüberwachung akttiviert wird über Jumper 18 Variable countAnzMaxSpanMin die minimale Spannung einstellen
   * 
   * <<<<<<<<<<<<<<<<<<<<<<!!!!!!!! Hinweise zur Software Inbetriebnahme !!!!!!!!>>>>>>>>>>>>>>>>>>>>>>
   * 
 * 
 * 
 * Getestet mit Sender/Empfänger
   * Graupner HOTT
   * Carson Multireflex
 * 

 * Änderungen:  
   *  
     *  06.03.2021 (V2.6)
     *  Diverse Code Korrekturen / Vermerk von zusätzlichen Hinweisen
     *  
     *  
   *     
   *  
     *  03.02.2021 (V2.5)
     *  Spannungsmessung einstellbar über Jumper auf Eingang A6 => GND keine Spannungsmessung
     *  Einlesen über AD Wandler der Spannung
   *    
   *  
     *  22.11.2020 (V2.4)
     *  Aktivierung über Jumper auf HW Board und Eingang A01 => Korrektur da A06 und A07 nur lesbar sind
     *  
   *    
   *  15.11.2020 (V2.3)
     *  Erweiterung mit 4 direkten analogen A2-A5 Eingangskanälen welche direkt auf die digitalen Ausgängen D10-D13 geleitet werden.
     *  Aktivierung über Jumper auf HW Board und Eingang A06
     *  KEIN PWM Signal Ausgang, sondern nur HIGH oder LOW Signal im Togglemodus
     *  Einstellung per Sofware in Setup Routine ob Tastfunktion oder selbsthalten pro Eingang => Standardmässig auf Tastfunktion (noch nicht erstell!!!!!)
   *    
   * 14.05.2020(V2.1)
     * Erweiterung auf 11 Ausgänge
     * Start LED Flackern auf Default Analog Pin A0 gesetzt
   *    
   * 14.05.2020(V2.1)
     * Erweiterung auf 11 Ausgänge
     * Start LED Flackern auf Default Analog Pin A0 gesetzt
   *     
 *  
 *  
 *  
 *<<<<<<<<<<<<<<<<<<<<<<!!!!!!!! Start Wichtige Hinweise (02.08.2020) !!!!!!!!>>>>>>>>>>>>>>>>>>>>>>
 * 
 * Test haben ergeben, das bei gewissen Servos das PWM Signal beim Steuern das Signal auf den Emfpägerausgang überspringt, 
 * welches zum MulitSwitch Print führt. Dieses Siganlüberspringen wird eingelesen und Zählt den Counter hoch.
 * Einzige Lösung bisher ist der Austausch des Servos.
 * Das Problem konnte mangels Alternativen an Fernsteuerungen/Empfängern nur mit GRAUPNER HOTT Anlagen getestet und nachvollzogen werden.
 * 
 * <<<<<<<<<<<<<<<<<<<<<<!!!!!!!! Ende wichtige Hinweise !!!!!!!!>>>>>>>>>>>>>>>>>>>>>>
 * 
 *  
 * 
 * 
 */




#define PinOut1 3
#define PinOut2 4
#define PinOut3 5
#define PinOut4 6
#define PinOut5 7
#define PinOut6 8
#define PinOut7 9
#define PinInOut8 10
#define PinInOut9 11
#define PinInOut10 12
#define PinInOut11 13

#define PinOutStart A0            //Blinkende LED im letzten Schritt der Setup Routine  => Ändern auf PIN ID A0 (Analog Output 0), ist die LED auf dem JoyIt Board
#define PinInStart1 A1
#define PinInStart2 A2
#define PinInStart3 A3
#define PinInStart4 A4
#define PinInStart5 A5
#define PinSetVin A6              //Spannungsmessung Aktivierung: GND inaktiv, VCC aktiv
#define PinInVin A7               //Ext. Spannungsmessung => z.B. eigenständiger Lipo Akku für Sonderfunktionen

const int Ausgaenge_Array[11] = {PinOut1, PinOut2, PinOut3, PinOut4, PinOut5, PinOut6, PinOut7, PinInOut8, PinInOut9, PinInOut10, PinInOut11};
volatile int Ausgaenge_Status[11] = {false, false, false, false, false, false, false, false, false, false, false};

volatile const int PinIn2 = 2;                  //Eingang für Counter Signal
unsigned long getPWM_Input=0;                   //Eingelesendes PWM Signal
volatile int counter = 0;                       //Counter für Ausgang Kanal schalten

//Variablen für MultiSwitch Funktionen
unsigned long aktZeit = 0;                            //Aktuelle Zeit
volatile unsigned long alteZeit = 0;                  //Zeit für die Deltaberechnung und Freigabe des Ausgangs
long delta = 0;                                       //Delta Zeit = aktZeit - alteZeit =>  muss grösser sein als die MultiSwitchZeit und der Ausgang wird geschaltet
volatile unsigned long MultiSwitchZeit = 1500;        //Zeit welche es keinen Wechsel mehr gibt und der index grösser 8 ist

//Zeit Variablen für Entprellen des PWM Signals
unsigned long aktZeit_PWM = 0;                                //Aktuelle Zeit
volatile unsigned long alteZeit_PWM = 0;                      //Delta Zeit für Freigabe
volatile unsigned long entPrell_PWM = 10;                     //Zeit für Prellen der Taster
volatile bool ErsterHighPuls = false;                         //Erster PWM Wert > 1800 oder <1200 (geht beides)
volatile bool FreigabeAusgang = false;                        //Freigabe für den Ausgang, erfolgt erst wenn Taster losgelassen wird

/*Hilfsvariablen setzen (Arraylänge) für Durchschaltung der Eingängen*/
/*Arrays für das Einlesen und setzten des Status der analogen Eingänge */
const int Eingang_An_Array[4] = {PinInStart2, PinInStart3, PinInStart4, PinInStart5};
volatile int Eingang_An_Status[4] = {false, false, false, false};
volatile int Eingang_An_Status_Freigabe[4] = {false, false, false, false};

volatile int maxCountArr = 11;              //Wenn A1 High ist (Direkte Schaltung Eingänge) wird beim Setup der Wert auf 7 gesetzt
volatile bool setDirEingaenge = false;
unsigned long setInput=0;                   //Eingelesendes PWM Signal für Schaltung des Ausgangs

/*Angaben Spannungsmessung*/
  bool setVin = false;            //A6 auf GND ist Spannungsmessung inaktiv

 
  int getSpannungInput;           //Eingelesener Wert an A7
  float setBerSpannung ;          //Wert mathematisch umwandeln um den Spannungswert in Volt zu erhalten
  int countAnzMinSpan = 0;        //Zählen wie häufig die Spannung unterschritten wird => nach 10 mal unteschreiten nacheinander wird Ausgang geschaltet und erst nach Neustart des Arduinos zurückgesetzt.
  int setSpanAus = false;         //Wert wird auf True gesetzt wenn die Spannung zu tief und der max. Counter Wert erreicht ist.
  
  float u_max = 0;                //Hilfsvariable für Berechnung der Spannung
  float u_reference = 0;          //Berechnete Refernzspannung
  long timealt = 0;               //Speicher für alte Zeit für Spannungsmessung






/*************************************************************************
 * Änderbare Variablen für die Spannungsmessung
 *************************************************************************/
 float setSpannungMin = 7.4;        //Minimale Spannung damit Augang ausgeschaltet wird
 int countAnzMaxSpanMin = 5;        //Counter Wert der überschritten werden muss damit der Ausgang schaltet. Counter schaltet jeweils zurück wenn Spannung nur kurzfristig unterschritten wird
 long TimeSpannungMessung = 10;     //Max. Wert für Counter Spannung, wenn erreicht wird Spannung überprüft
 int R26 = 992;                     //Widerstand 1000 Ohm => Widerstand möglichst genau ausmessen!
 int R27 = 100;                      //Widerstand 100 Ohm => Widerstand möglichst genau ausmessen!








void setup() {

   //Serial.begin(115200);
   //Serial.println("<<<<<<<<<<<<<<<<<<<< Start Setup Routine! >>>>>>>>>>>>>>>>>>>>>>");

   //Aktivierung Spannungsüberwachung => Input auf GND ist Wert 0
  if(analogRead(PinSetVin) > 10)
    setVin = true;
  
  //Eingänge Initialisieren  
  pinMode(PinIn2, INPUT_PULLUP);
  digitalWrite(PinIn2, LOW);   
  pinMode(PinInStart1, INPUT_PULLUP);
  digitalWrite(PinInStart1, LOW);   

  //Wenn High, digitale Durchschaltung der Eingänge D10-D13
  if(digitalRead(PinInStart1) == HIGH)
  {
    maxCountArr = 7;
    setDirEingaenge = true;

    //Analoge Pins als Ausgänge definierne
    for(int i = 0; i < 4; i++)
    {
      pinMode(Eingang_An_Array[i], INPUT_PULLUP);
      digitalWrite(Eingang_An_Array[i], LOW);
    }

      //Definieren welche Eingänge aktiv (PWM Signal am Pin) sind, nur diese werden später eingelesen
      //Wert 4 ist Status inaktiv und am Eingang kein Signal anliegend
      for(int xx = 0; xx < 4; xx++)
      {
        if(pulseIn(Eingang_An_Array[xx], HIGH) > 10)      //PWM Signal liegt an
          Eingang_An_Status_Freigabe[xx] = true;
        else
          Eingang_An_Status_Freigabe[xx] = false;            
      }
      
  }



 /***********************************************************************************
 Wenn Digitale Eingänge durchgeschaltet (PinInStart1 == High) und Spannungsmessung Aktiv (setVin == true)
 Pin D9 für Relais Schaltung (Abfall Spannungsmessung) aktivert ist maxCountArr auf 6 redziert
 Wenn nur Spannungsmessung aktiv => maxCountArr auf 10 reduziert und Pin D13 für Relais Schaltung (Abfall Spannungsmessung)
 ***********************************************************************************/
  if(digitalRead(PinInStart1) == HIGH && setVin == true)
    maxCountArr = 6;
  if(digitalRead(PinInStart1) == LOW && setVin == true)
    maxCountArr = 10;  

  
  //Ausgänge Initialisieren
  for(int i = 0; i < 11; i++)
  {
    pinMode(Ausgaenge_Array[i], OUTPUT);
  }


  //Ausgang für Sichtbarkeit das Programm gestartet ist
  pinMode(PinOutStart, OUTPUT);
  
  /*****************************************************************************
  Start Kontrollausgabe nach dem Setup
  *******************************************************************************/
  
  digitalWrite(PinOutStart, HIGH);
  delay(1000);
  digitalWrite(PinOutStart, LOW);
  delay(1000);
  
  for(int i = 0; i < 5; i++)
  {
    digitalWrite(PinOutStart, HIGH);
    delay(200); 
    digitalWrite(PinOutStart, LOW);
    delay(200);
  }

  digitalWrite(PinOutStart, HIGH);
  delay(2000);
  digitalWrite(PinOutStart, LOW);

  
  
  //Ausgänge auf 0 setzen
  for(int i = 0; i < 11; i++)
  {
    digitalWrite(Ausgaenge_Array[i], LOW);
  } 

 
  /*****************************************************************************
  Ende Kontrollausgabe nach dem Setup
  *******************************************************************************/  
  
   //Serial.println("<<<<<<<<<<<<<<<<<<<< Ende Setup Routine! >>>>>>>>>>>>>>>>>>>>>>");
  
}





void loop() {

  //PWM Signal ab Empfänger einlesen
  getPWM_Input = pulseIn(PinIn2, HIGH);


  //Spannungsmessung wenn aktiv Pin A6
  if(setVin == true)
  {
      if((millis()- timealt) > (TimeSpannungMessung*1000))
      {
        ReadSpannung();       //Spannung messen       
        timealt = millis();
      }    
  }


  
  /**********************************************************************************
   * Aufruf der Taster Funktion für das Zählen des Ausgangs
   ***********************************************************************************/
  if((getPWM_Input < 1200 || getPWM_Input > 1800) && ErsterHighPuls == false)
    getCounter();


  

  /***********************************************************************************
   * Der Taster kann so lange gedrückt bleiben, es wir nur immer 1 mal hochgezählt
   * Ist das PWM Signal in Ruhemodus, kann weiter hochgezählt werden oder die Freigabe für den Ausgang erfolgt
   ***********************************************************************************/
  if((getPWM_Input < 1200 || getPWM_Input > 1800) && ErsterHighPuls == true)
      FreigabeAusgang = false;
  else if(getPWM_Input > 1200 && getPWM_Input < 1800 && ErsterHighPuls == true && alteZeit_PWM > 0)// && counter > 0)
  {
    //alteZeit_PWM = 0;
    ErsterHighPuls = false;
    FreigabeAusgang = true;
    
  }


  
  
  /***********************************************************************************
   * Aufruf Ausgang Schalt Funktion wenn der Taster bedient wurde
   ***********************************************************************************/
  if(counter > 0 && FreigabeAusgang == true &&   (millis()- alteZeit) >  MultiSwitchZeit  )
    setAusgang();



  /************************************************************************************
  Einlesen D10-D13 bei direkter Durchschaltung
  ************************************************************************************/
  if(setDirEingaenge == true)
    setDirAusgang();



}



/*****************************************************************************************
Durchschalten der analogen Eingänge wenn A06 aktiv
*******************************************************************************************/
void setDirAusgang()
{

      int counttemp = 7;
      
      for(int i = 0; i < 4; i++)
      {
        if(Eingang_An_Status_Freigabe[i] == true) 
        {
            //Einlesen des Signal
            setInput = pulseIn(Eingang_An_Array[i], HIGH); 
           
            if(setInput < 1200 || setInput > 1800 && Eingang_An_Status[i] == false)
            {
                digitalWrite(Ausgaenge_Array[counttemp], HIGH);
                Eingang_An_Status[i] = true;
            }
            
            if(setInput > 1200 && setInput < 1800 && Eingang_An_Status[i] == true)
            {
                digitalWrite(Ausgaenge_Array[counttemp], LOW); 
                Eingang_An_Status[i] = false;

            }

           
        }
        
         counttemp = counttemp + 1;
      }

}


/*****************************************************************************************
 Einlesen des PWM Signals am Pin 2
*******************************************************************************************/
void getCounter()
{

  if((millis() - alteZeit_PWM) > entPrell_PWM && alteZeit_PWM > 0)
  {
      alteZeit = alteZeit_PWM = millis();
      
      //Counter erhöhen, ist die max. Ausgangsbelegung erreicht beginnt der Counter wieder von vorne
      if(counter == 11)
        counter = 1;
      else
        counter = counter + 1;
;
      //Erster Puls ist durch, index wurde hochgezählt
      ErsterHighPuls = true;

  }
  else
    alteZeit_PWM = millis();    //Setzen der aktuellen zeit für die Berechnung der Freigabe des Ausgangs

}



/********************************************************************************
Schalten des Ausganzg gemäss Counter Stand durch das hochzählen des Einganges
*********************************************************************************/
void setAusgang()
{
        //Array beginnt bei 0 => Counter von 1 (Einfach zu zählen...)
        int arrind = counter-1;
        
        //Inkrementierung des Ausgangsignals
        if(Ausgaenge_Status[arrind] == false)
        {
          digitalWrite(Ausgaenge_Array[arrind], HIGH);
          Ausgaenge_Status[arrind] = true;
        }
        else
        {
          digitalWrite(Ausgaenge_Array[arrind], LOW);
          Ausgaenge_Status[arrind] = false;      
        }
  
        //Zurücksetzen der Index, damit wieder von vorne durchgezählt werden kann
        counter = 0;
        alteZeit = 0;
        alteZeit_PWM = 0;

}


/********************************************************************************
Spannungsmessung: Interne Referenzspannung
*********************************************************************************/
long readVcc() {
  long result;
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2);
  ADCSRA |= _BV(ADSC);
  while (bit_is_set(ADCSRA,ADSC));
  result = ADCL;
  result |= ADCH<<8;
  result = 1126400L / result;
  return result;
}


/********************************************************************************
Spannungsmessung
*********************************************************************************/
void ReadSpannung()
{
  u_reference = readVcc() / 1000.0;
  u_max = u_reference*(R26+R27) / R27;
  
  setBerSpannung = u_max/1024*analogRead(PinInVin);

      if(setBerSpannung < setSpannungMin)
        countAnzMinSpan++;
      else
        countAnzMinSpan = 0;

     if(countAnzMinSpan > countAnzMaxSpanMin && setSpanAus == false)
     {
    
        if(maxCountArr == 6)  //Eingänge durchgeschaltet
          digitalWrite(Ausgaenge_Array[6], HIGH);
    
        if(maxCountArr == 10) //Nur Taster Funktion
          digitalWrite(Ausgaenge_Array[10], HIGH);
          
        setSpanAus = true;
     }
}
